package server

import (
	"bytes"
	"encoding/json"
	"log"
	"os"
	"strconv"
	"time"

	redis "github.com/go-redis/redis/v8"
	"github.com/jackc/pgx/v5/pgxpool"
	"github.com/streadway/amqp"

	"git.pbiernat.dev/egommerce/go-api-pkg/consul"
	"git.pbiernat.dev/egommerce/go-api-pkg/fluentd"

	db "git.pbiernat.dev/egommerce/basket-service/pkg/database"
	p "git.pbiernat.dev/egommerce/basket-service/pkg/server"

	cnf "git.pbiernat.dev/egommerce/basket-service/internal/config"
)

type (
	Server struct {
		Base     *p.Server
		Config   *cnf.Config
		Cache    *redis.Client
		Database *pgxpool.Pool
		Eventbus *amqp.Channel
		Logger   *fluentd.Logger
		Registry *consul.Service
	}
	OptionFn func(*Server) error // FIXME: similar in worker
)

func New(c *cnf.Config, opts ...OptionFn) *Server {
	svr := &Server{
		Base:   p.New(c.Base),
		Config: c,
	}

	for _, opt := range opts {
		if err := opt(svr); err != nil {
			log.Fatalf("Failed to attach extension to the server. Err: %v\n", err)
		}
	}

	svr.Base.ShutdownFn = svr.Shutdown()

	SetupMiddleware(svr)
	SetupRouter(svr)

	return svr
}

func (s *Server) Shutdown() p.ShutdownFn {
	return func() error {
		s.Logger.Log("Server %s is going down...", s.Base.AppID)

		s.Registry.Unregister()
		// s.clearMetadataCache()
		s.Eventbus.Close()
		s.Database.Close()
		s.Logger.Log("Gone.")
		s.Logger.Close()

		return s.Base.Shutdown()
	}
}

// @CHECK: merge s.Config and s.Base.Config to display all config as one array/map
func (s *Server) registerKVUpdater() { // @FIXME: merge duplication in server.go and worker.go
	go func() {
		ticker := time.NewTicker(time.Second * 10)
		for range ticker.C {
			config, _, err := s.Registry.KV().Get(s.Config.KVNamespace, nil)
			if err != nil || config == nil {
				return
			}

			kvCnf := bytes.NewBuffer(config.Value)
			decoder := json.NewDecoder(kvCnf)
			if err := decoder.Decode(&s.Config); err != nil {
				return
			}
		}
	}()
}

// func (s *Server) clearMetadataCache() {
// 	ctx := context.Background()
// 	key, address := s.getMetadataIPsKey(), s.Config.Base.AppID

// 	s.Cache.LRem(ctx, key, 0, address)
// }

// func (s *Server) getMetadataIPsKey() string {
// 	return "internal__" + s.Base.Config.AppName + "__ips"
// }

func WithCache(c *cnf.Config) OptionFn {
	return func(s *Server) error {
		s.Cache = redis.NewClient(&redis.Options{
			Addr:     s.Config.CacheAddr,
			Password: s.Config.CachePassword,
			DB:       0,
		})

		return nil
	}
}

func WithDatabase(c *cnf.Config) OptionFn {
	dbConn, err := db.Connect(c.DbURL)
	if err != nil {
		log.Fatalf("Failed to connect to the Database: %s. Err: %v\n", c.DbURL, err)
		os.Exit(1)
	}

	return func(s *Server) error {
		s.Database = dbConn

		return nil
	}
}

func WithEventbus(c *cnf.Config) OptionFn {
	return func(s *Server) error {
		conn, err := amqp.Dial(s.Config.EventBusURL)
		if err != nil {
			log.Fatalf("Failed to connect to the Eventbus: %s. Err: %v\n", s.Config.EventBusURL, err)
		}

		chn, err := conn.Channel()
		if err != nil {
			log.Fatalf("Failed to open new Eventbus channel. Err: %v\n", err)
		}

		s.Eventbus = chn

		return nil
	}
}

func WithLogger(c *cnf.Config) OptionFn {
	return func(s *Server) error {
		logHost, logPort, err := fluentd.ParseAddr(c.LoggerAddr)
		if err != nil {
			log.Fatalf("Failed to parse Fluentd address: %s. Err: %v", c.LoggerAddr, err)
		}

		logger, err := fluentd.NewLogger(c.Base.GetAppFullName(), logHost, logPort)
		if err != nil {
			log.Fatalf("Failed to connect to the Fluentd on %s:%d. Err: %v", logHost, logPort, err)
		}

		s.Logger = logger

		return nil
	}
}

func WithRegistry(c *cnf.Config) OptionFn {
	return func(s *Server) error {
		port, _ := strconv.Atoi(c.Base.NetAddr[1:]) // FIXME: can be IP:PORT which will cause error
		log.Printf("Consul retrieved port: %v", port)
		registry, err := consul.NewService(c.RegistryAddr, c.Base.AppID, c.Base.AppName, c.Base.GetIP(), c.Base.AppDomain, c.Base.PathPrefix, port)
		if err != nil {
			log.Fatalf("Failed to connect to the Consul on: %s. Err: %v", c.RegistryAddr, err)
		}

		err = registry.Register()
		if err != nil {
			log.Fatalf("Failed to register in the Consul service. Err: %v", err)
		}

		registry.RegisterHealthChecks()
		s.registerKVUpdater()

		s.Registry = registry

		// svc, _ := registry.Connect()
		// tlsCnf := svc.ServerTLSConfig()
		// s.Base.App.Server().TLSConfig = tlsCnf
		// fmt.Println("Podmiana configa TLS")
		// defer svc.Close()

		// go func() { // Consul KV updater
		// 	ticker := time.NewTicker(time.Second * 15)
		// 	for range ticker.C {
		// 		fetchKVConfig(s) // FIXME: duplicated in worker
		// 	}
		// }()

		// go func() { // Server metadata cache updater
		// 	ticker := time.NewTicker(time.Second * 5)
		// 	for range ticker.C {
		// 		s.cacheMetadata()
		// 	}
		// }()

		return nil
	}
}
